from pydantic import BaseModel, Field
from typing import Union, List, Any, Dict
from ._stage_type import StageType


class BaseTable(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    stage: StageType = Field(...)
    database: str = Field(...)
    name: str = Field(...)
    id: Union[str, List[str]] = Field(default=[])


class Table(BaseTable):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    depends_on: List[str] = Field(default=[])
    table_properties: Dict[str, str] = Field(default=None)


class TableMapping(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    destination: Table = Field(...)
    source: Union[Dict[str, Table], Table] = Field(...)



from ._deltalake import DeltaLake
from ._read import Read
from ._dataset import DataSet
from ._factory import factory as dataset_factory

__all__ = ["DeltaLake", "Read", "dataset_factory", "DataSet"]



import logging
from pydantic import BaseModel, Field, PrivateAttr
from .._utils import JinjaVariables
from typing import Any, Dict
from .._timeslice import Timeslice
from .._stage_type import StageType


class DataSet(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._logger = logging.getLogger(self.__class__.__name__)

    _logger: Any = PrivateAttr(default=None)
    _replacements: Dict[JinjaVariables, str] = PrivateAttr(default=None)
    database: str = Field(...)
    destination_table: str = Field(...)
    table: str = Field(...)
    container: str = Field(...)
    root: str = Field(...)
    path: str = Field(default=None)
    options: dict = Field(...)
    timeslice: Timeslice = Field(...)
    checkpoint: str = Field(default=None)
    stage: StageType = Field(...)

    def _render(self):
        pass



from enum import Enum
from ._deltalake import DeltaLake
from ._read import Read
from ._write import Write
from ._dataset import DataSet
import logging
from .._stage_type import StageType
from .._table import Table
from .._timeslice import Timeslice
from typing import Union


class IOType(Enum):
    READ = "read"
    DELTALAKE = "deltalake"
    WRITE = "write"


class _DatasetFactory:
    _TIMESLICE = "timeslice"
    _TABLE = "destination_table"
    _TABLE_PROPERTIES = "table_properties"
    _STAGE = "stage"
    _DATABASE = "database"

    def __init__(self) -> None:
        self._logger = logging.getLogger(self.__class__.__name__)
        self._dataset = {}

    def register_dataset_type(self, io_type: StageType, dataset_type: type):
        self._logger.debug(f"Register dataset type {dataset_type} as {type}")
        self._dataset[io_type] = dataset_type

    def _get_dataset_type(
        self,
        stage: StageType,
        dataset_config: dict,
    ) -> DataSet:
        type: IOType = IOType(stage.value.lower())

        self._logger.debug(f"Get {type.name} from factory dataset")
        dataset_class = self._dataset.get(type)

        if not dataset_class:
            self._logger.debug(
                f"IOType {type.name} not registered in the dataset factory dataset"
            )
            raise ValueError(type)

        return dataset_class(
            **dataset_config,
        )

    def get_data_set(
        self, config: dict, dataset: Union[Table, dict], timeslice: Timeslice
    ):
        if isinstance(dataset, dict):
            for name, table_obj in dataset.items():
                stage_config = self._get_stage_table_config(
                    config, table_obj, timeslice
                )
                dataset[name] = self._get_dataset_type(table_obj.stage, stage_config)

        elif isinstance(dataset, Table):
            stage_config = self._get_stage_table_config(config, dataset, timeslice)
            dataset = self._get_dataset_type(dataset.stage, stage_config)

        return dataset

    def _get_stage_table_config(self, config: dict, table: Table, timeslice: Timeslice):
        stage_config = config[table.stage.name]
        stage_config[self._TIMESLICE] = timeslice
        stage_config[self._TABLE] = table.name
        stage_config[self._TABLE_PROPERTIES] = table.table_properties
        stage_config[self._STAGE] = table.stage
        stage_config[self._DATABASE] = table.database

        return stage_config


factory = _DatasetFactory()
factory.register_dataset_type(IOType.READ, Read)
factory.register_dataset_type(IOType.DELTALAKE, DeltaLake)
factory.register_dataset_type(IOType.WRITE, Write)



from datetime import datetime
from typing import Literal, Union
from pydantic import BaseModel, Field
from typing import Any

_WILDCARD = "*"
Wildcard = Literal["*"]

_UNSUPPORTED_FORMAT_CODES = [
    "%c",
    "%x",
    "%X",
    "%G",
    "%u",
    "%V",
    "%z",
    "%Z",
    "%I",
    "%p",
    "%b",
    "%B",
    "%a",
    "%A",
    "%w",
]


class Timeslice(BaseModel):
    def __init__(__pydantic_self__, **data: Any) -> None:
        super().__init__(**data)

    year: Union[int, Wildcard] = Field(...)
    month: Union[int, Wildcard] = Field(default=_WILDCARD)
    day: Union[int, Wildcard] = Field(default=_WILDCARD)
    hour: Union[int, Wildcard] = Field(default=0)
    minute: Union[int, Wildcard] = Field(default=0)
    second: Union[int, Wildcard] = Field(default=0)
    microsecond: Union[int, Wildcard] = Field(default=0)

    def strftime(self, format: str):
        """This will format and return the timeslice using python format codes. Only a subset of format codes are suppoered by design
        %d - Day of the month as a zero-padded decimal number.
        %m - Month as a zero-padded decimal number.
        %y - Year without century as a zero-padded decimal number.
        %Y - Year with century as a decimal number.
        %H - Hour (24-hour clock) as a zero-padded decimal number.
        %M - Minute as a zero-padded decimal number.
        %S - Second as a zero-padded decimal number.
        %f - Microsecond as a decimal number, zero-padded to 6 digits.
        %% - A literal '%' character.
        %j - Day of the year as a zero-padded decimal number.
        %U - Week number of the year (Sunday as the first day of the week) as a zero-padded decimal number. All days in a new year preceding the first Sunday are considered to be in week 0.
        %W - Week number of the year (Monday as the first day of the week) as a zero-padded decimal number. All days in a new year preceding the first Monday are considered to be in week 0.
        NOT SUPPORTED - %c - Locale’s appropriate date and time representation.
        NOT SUPPORTED - %x - Locale’s appropriate date representation.
        NOT SUPPORTED - %X - Locale’s appropriate time representation.
        NOT SUPPORTED - %G - ISO 8601 year with century representing the year that contains the greater part of the ISO week (%V).
        NOT SUPPORTED - %u - ISO 8601 weekday as a decimal number where 1 is Monday.
        NOT SUPPORTED - %V - ISO 8601 week as a decimal number with Monday as the first day of the week. Week 01 is the week containing Jan 4.
        NOT SUPPORTED - %z - UTC offset in the form ±HHMM[SS[.ffffff]] (empty string if the object is naive).
        NOT SUPPORTED - %Z - Time zone name (empty string if the object is naive).
        NOT SUPPORTED - %I - Hour (12-hour clock) as a zero-padded decimal number.
        NOT SUPPORTED - %p - Locale’s equivalent of either AM or PM.
        NOT SUPPORTED - %b - Month as locale’s abbreviated name.
        NOT SUPPORTED - %B - Month as locale’s full name.
        NOT SUPPORTED - %a - Weekday as locale’s abbreviated name.
        NOT SUPPORTED - %A - Weekday as locale’s full name.
        NOT SUPPORTED - %w - Weekday as a decimal number, where 0 is Sunday and 6 is Saturday.
        """

        unsupported_codes = [c for c in _UNSUPPORTED_FORMAT_CODES if c in format]

        if unsupported_codes:
            unsupported_codes = ",".join(unsupported_codes)
            raise Exception(
                f"The format contains the following unsupported format codes: {unsupported_codes}"
            )

        format, _year = self._format_wildcard(format, self.year, ["%y", "%Y"], 1900)
        format, _month = self._format_wildcard(format, self.month, "%m", 1)
        format, _day = self._format_wildcard(format, self.day, "%d", 1)

        format, _hour = self._format_wildcard(format, self.hour, "%H")
        format, _minutue = self._format_wildcard(format, self.minute, "%M")
        format, _second = self._format_wildcard(format, self.second, "%S")
        format, _microsecond = self._format_wildcard(format, self.microsecond, "%f")

        timeslice = datetime(
            _year, _month, _day, _hour, _minutue, _second, _microsecond
        )

        formatted = timeslice.strftime(format)
        return formatted

    def _format_wildcard(
        self,
        format: str,
        datepart: Union[int, Wildcard],
        format_code: Union[list, str],
        default=0,
    ):
        if datepart == _WILDCARD:
            if isinstance(format_code, str):
                format = format.replace(format_code, f"{_WILDCARD}")
            elif isinstance(format_code, list):
                for f in format_code:
                    format = format.replace(f, f"{_WILDCARD}")
            datepart = default

        return format, datepart

    def __str__(self) -> str:
        return self.strftime("%Y-%m-%d %H:%M:%S.%f")


class TimesliceNow(Timeslice):
    def __init__(self) -> None:
        now = datetime.now()
        args = {
            "year": now.year,
            "month": now.month,
            "day": now.day,
            "hour": now.hour,
            "minute": now.minute,
            "second": now.second,
            "microsecond": now.microsecond,
        }
        super().__init__(**args)


class TimesliceUtcNow(Timeslice):
    def __init__(self) -> None:
        now = datetime.utcnow()
        args = {
            "year": now.year,
            "month": now.month,
            "day": now.day,
            "hour": now.hour,
            "minute": now.minute,
            "second": now.second,
            "microsecond": now.microsecond,
        }
        super().__init__(**args)



import yaml
import os
from .dataset import DataSet
from ._timeslice import Timeslice
from ._tables import Tables, _INDEX_WILDCARD
from ._stage_type import StageType
from .dataset import dataset_factory


class Config:
    _CONFIG_PATH = "../Config/"
    _CONFIG_FILE = "config.yaml"
    _ENCODING = "utf-8"
    _TABLES = "tables"
    _SOURCE_TABLE = "source_table"

    def __init__(self, config_path: str = None):
        self.config = {}

        if not config_path:
            config_path = os.path.join(self._CONFIG_PATH, self._CONFIG_FILE)

        with open(config_path, "r", encoding=self._ENCODING) as f:
            self.config = yaml.safe_load(f)

        _tables_path = self.config["tables"]

        with open(_tables_path, "r", encoding=self._ENCODING) as f:
            self.config["tables"] = yaml.safe_load(f)

        self.tables = Tables(table_data=self.config["tables"])

    def get_table_mapping(
        self,
        timeslice: Timeslice,
        stage: StageType,
        table: str = _INDEX_WILDCARD,
        database: str = _INDEX_WILDCARD,
        index: str = None,
    ):
        table_mapping = self.tables.get_table_mapping(
            stage=stage, table=table, database=database, index=index
        )

        table_mapping.source = dataset_factory.get_data_set(
            self.config, table_mapping.source, timeslice
        )
        table_mapping.destination = dataset_factory.get_data_set(
            self.config, table_mapping.destination, timeslice
        )

        return table_mapping

    def link_checkpoint(
        self,
        source: DataSet,
        destination: DataSet,
        checkpoint_name: str = None,
    ):
        if not checkpoint_name:
            checkpoint_name = f"{source.database}.{source.table}-{destination.database}.{destination.table}"

        source.checkpoint = checkpoint_name
        source._render()
        destination.checkpoint = checkpoint_name
        destination._render()
